package cn.com.qmth.stmms.biz.mark.service.Impl;

import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import cn.com.qmth.stmms.biz.config.service.SystemConfigService;
import cn.com.qmth.stmms.biz.config.service.impl.SystemCache;
import cn.com.qmth.stmms.biz.exam.model.Exam;
import cn.com.qmth.stmms.biz.exam.model.ExamStudent;
import cn.com.qmth.stmms.biz.exam.model.ExamSubject;
import cn.com.qmth.stmms.biz.exam.model.MarkGroup;
import cn.com.qmth.stmms.biz.exam.model.Marker;
import cn.com.qmth.stmms.biz.exam.service.ExamService;
import cn.com.qmth.stmms.biz.exam.service.ExamStudentService;
import cn.com.qmth.stmms.biz.exam.service.ExamSubjectService;
import cn.com.qmth.stmms.biz.exam.service.MarkGroupService;
import cn.com.qmth.stmms.biz.exam.service.MarkerService;
import cn.com.qmth.stmms.biz.lock.LockService;
import cn.com.qmth.stmms.biz.mark.service.MarkService;
import cn.com.qmth.stmms.biz.report.service.ReportService;
import cn.com.qmth.stmms.biz.utils.TaskLockUtil;
import cn.com.qmth.stmms.common.enums.ConfigType;
import cn.com.qmth.stmms.common.enums.ExamStatus;
import cn.com.qmth.stmms.common.enums.LockType;
import cn.com.qmth.stmms.common.enums.MarkStatus;
import cn.com.qmth.stmms.common.enums.SubjectiveStatus;

/**
 * 与评卷相关的所有定时任务
 *
 * @author luoshi
 */
@Component
public class MarkCronService {

    protected static final Logger log = LoggerFactory.getLogger(MarkCronService.class);

    @Autowired
    private ExamStudentService studentService;

    @Autowired
    private ExamSubjectService subjectService;

    @Autowired
    private MarkGroupService groupService;

    @Autowired
    private MarkService markService;

    @Autowired
    private LockService lockService;

    @Autowired
    private MarkerService markerService;

    @Autowired
    private SystemCache systemCache;

    @Autowired
    private ExamService examService;

    @Value("${mark.activeExpireMinute}")
    private long markerActiveExpireMinute;

    @Autowired
    private ReportService reportService;

    @Autowired
    private SystemConfigService systemConfigService;

    /**
     * 自动释放可清除的锁
     */
    @Scheduled(cron = "${mark.cleanLockSchedule}")
    public void cronCleanLock() {
        try {
            lockService.clear();
        } catch (Exception e) {
            log.error("CronCleanTask error", e);
        }
    }

    /**
     * 自动释放超时未处理的评卷任务
     */
    @Scheduled(cron = "${mark.cleanTaskSchedule}")
    public void cronCleanTask() {
        try {
            long timeoutMinute = systemCache.getMarkTime();
            TaskLockUtil.clearTimeoutTask(timeoutMinute * 60 * 1000);
        } catch (Exception e) {
            log.error("CronCleanTask error", e);
        }
    }

    /**
     * 定时生成评卷任务
     */
    @Scheduled(fixedDelay = 60 * 1000, initialDelay = 60 * 1000)
    public void buildLibrary() {
        log.info("start auto-create library");
        try {
            List<Integer> examIds = groupService.findExamIdByStatus(MarkStatus.TRIAL, MarkStatus.FORMAL);
            for (Integer examId : examIds) {
                Exam exam = examService.findById(examId);
                if (ExamStatus.FINISH.equals(exam.getStatus())) {
                    continue;
                }
                buildLibraryByExam(examId);
            }
        } catch (Exception e) {
            log.error("auto-create library error", e);
        } finally {
            log.info("finish auto-create library");
        }
    }

    /**
     * 定时更新评卷员质量监控指标
     */
    @Scheduled(fixedDelay = 60 * 1000, initialDelay = 60 * 1000)
    public void updateMarkerQuality() {
        log.info("start auto-update marker");
        try {
            List<Integer> examIds = groupService.findExamIdByStatus(MarkStatus.FORMAL);
            for (Integer examId : examIds) {
                updateMarkerQualityByExam(examId);
            }
        } catch (Exception e) {
            log.error("auto-update marker error", e);
        } finally {
            log.info("finish auto-update marker");
        }
    }

    private void updateMarkerQualityByExam(Integer examId) {
        // 获取主观题总分大于0的科目
        List<ExamSubject> subjects = subjectService.list(examId, 0);
        for (ExamSubject subject : subjects) {
            List<MarkGroup> groups = groupService.findByExamAndSubjectAndStatus(examId, subject.getCode(),
                    MarkStatus.FORMAL);
            for (MarkGroup markGroup : groups) {
                List<Marker> markers = markerService.findByExamAndSubjectAndGroup(examId, subject.getCode(),
                        markGroup.getNumber());
                for (Marker marker : markers) {
                    try {
                        if (markService.needUpdateQuality(marker, markerActiveExpireMinute)) {
                            log.info("start auto-update markerId: " + marker.getId());
                            markService.updateQuality(marker);
                        }
                    } catch (Exception e) {
                        log.error("update marker quality error for markerId=" + marker.getId(), e);
                    }
                }
            }
        }
    }

    private void buildLibraryByExam(Integer examId) {
        // 获取主观题总分大于0的科目
        List<ExamSubject> subjects = subjectService.listSubjectiveScore(examId, 0d);
        for (ExamSubject subject : subjects) {
            // 清除缺考考生和违纪考生
            List<ExamStudent> list = studentService.findAbsentOrBreachLibraryStudent(subject.getExamId(),
                    subject.getCode());
            if (list != null) {
                for (ExamStudent student : list) {
                    try {
                        lockService.waitlock(LockType.STUDENT, student.getId());
                        markService.deleteByStudent(student);
                    } catch (Exception e) {
                        log.error("delete student library error", e);
                    } finally {
                        lockService.unlock(LockType.STUDENT, student.getId());
                    }
                }
            }
            // 处理正常考生
            // 生成试评任务
            List<MarkGroup> groups = groupService.findByExamAndSubjectAndStatus(subject.getExamId(), subject.getCode(),
                    MarkStatus.TRIAL);
            long trialCount = studentService.countByExamIdAndSubjectCodeAndStatus(examId, subject.getCode(),
                    SubjectiveStatus.TRIAL, SubjectiveStatus.MARKED);
            if (!groups.isEmpty() && trialCount < subject.getTrialCount()) {
                while (trialCount < subject.getTrialCount()) {
                    // 取一个的考生
                    ExamStudent student = studentService.randomStudent(examId, subject.getCode());
                    if (student == null) {
                        break;
                    }
                    studentService.updateSubjectiveStatusAndTimeAndInspectorId(student.getId(), SubjectiveStatus.TRIAL,
                            null, null);
                    trialCount++;
                }
            }
            for (MarkGroup group : groups) {
                buildTrialLibrary(trialCount, group);
            }
            // 生成正评任务
            // 试评状态的分组也提前生成
            groups = groupService.findByExamAndSubjectAndStatus(subject.getExamId(), subject.getCode(),
                    MarkStatus.FORMAL, MarkStatus.TRIAL);
            for (MarkGroup group : groups) {
                buildFormalLibrary(subject, group);
            }
        }
    }

    private void buildFormalLibrary(ExamSubject subject, MarkGroup group) {
        Date lastBuildTime = group.getBuildTime();
        int count = 0;
        ExamStudent student = studentService.findUnLibraryStudent(group.getExamId(), group.getSubjectCode(),
                group.getNumber(), lastBuildTime);
        while (student != null) {
            // 尝试构造评卷任务
            try {
                lockService.watch(LockType.GROUP, group.getExamId(), group.getSubjectCode(), group.getNumber());
                lockService.watch(LockType.STUDENT, student.getId());
                // 上锁后重复验证分组状态
                if (!groupService.validateStatus(group.getExamId(), group.getSubjectCode(), group.getNumber(),
                        MarkStatus.FORMAL, MarkStatus.TRIAL)) {
                    break;
                }
                // 上锁后重复验证考生状态
                if (studentService.validateStatus(student.getId())) {
                    markService.buildFormalLibrary(student, group, subject);
                    lastBuildTime = student.getUploadTime();
                    count++;
                }
            } catch (Exception e) {
                log.error(
                        "build formal library error for studentId=" + student.getId() + ", groupNumber="
                                + group.getNumber(), e);
            } finally {
                lockService.unwatch(LockType.STUDENT, student.getId());
                lockService.unwatch(LockType.GROUP, group.getExamId(), group.getSubjectCode(), group.getNumber());
            }
            // 取下一个考生
            student = studentService.findUnLibraryStudent(group.getExamId(), group.getSubjectCode(), group.getNumber(),
                    lastBuildTime);
        }
        // 有新任务创建，同时正评状态，此时才需要更新任务数量
        if (count > 0) {
            group = groupService.findOne(group.getExamId(), group.getSubjectCode(), group.getNumber());
            if (group.getStatus() == MarkStatus.FORMAL) {
                markService.updateLibraryCount(group);
                markService.updateMarkedCount(group);
            }
        }
    }

    private void buildTrialLibrary(long trialCount, MarkGroup group) {
        while (trialCount > group.getLibraryCount()) {
            // 取一个未生成试评任务的考生
            ExamStudent student = studentService.randomUnTrialStudent(group.getExamId(), group.getSubjectCode(),
                    group.getNumber());
            if (student == null) {
                return;
            }
            // 尝试构造试评任务
            try {
                lockService.watch(LockType.GROUP, group.getExamId(), group.getSubjectCode(), group.getNumber());
                lockService.watch(LockType.STUDENT, student.getId());
                // 上锁后重新验证分组状态
                if (!groupService.validateStatus(group.getExamId(), group.getSubjectCode(), group.getNumber(),
                        MarkStatus.TRIAL)) {
                    break;
                }
                // 上锁后重复验证考生状态
                if (studentService.validateStatus(student.getId())) {
                    markService.buildTrialLibrary(student, group);
                    markService.updateLibraryCount(group);
                    markService.updateMarkedCount(group);
                }
            } catch (Exception e) {
                log.error(
                        "build trial library error for studentId=" + student.getId() + ", groupNumber="
                                + group.getNumber(), e);
            } finally {
                lockService.unwatch(LockType.STUDENT, student.getId());
                lockService.unwatch(LockType.GROUP, group.getExamId(), group.getSubjectCode(), group.getNumber());
            }
        }
    }

    @Scheduled(fixedDelay = 30 * 60 * 1000, initialDelay = 2 * 60 * 1000)
    public void updateReport() {
        String systemConfig = systemConfigService.findByType(ConfigType.AUTO_REPORT);
        if (systemConfig == null || !systemConfig.equalsIgnoreCase("on")) {
            return;
        }
        log.info("start auto-update report");
        try {
            Map<Integer, Set<String>> map = new HashMap<Integer, Set<String>>();
            List<Integer> examIds = groupService.findExamIdByStatus(MarkStatus.FORMAL);
            for (Integer examId : examIds) {
                List<ExamSubject> list = subjectService.list(examId);
                Set<String> set = new HashSet<>();
                for (ExamSubject subject : list) {
                    if (!lockService.isLocked(LockType.SCORE_CALCULATE, subject.getExamId(), subject.getCode())) {
                        set.add(subject.getCode());
                    }
                }
                if (!set.isEmpty()) {
                    map.put(examId, set);
                }
            }
            for (Integer examId : map.keySet()) {
                reportService.reportExam(examId, map.get(examId));
            }
        } catch (Exception e) {
            log.error("auto-update report error", e);
        } finally {
            log.info("finish auto-update report");
        }
    }

}
